"""
files.py: IDE file access functions.
"""
# TODO: Progress callback for `get_doc()` (separate from the `callback` argument
#  of `idelib.importer.openFile()` and `idelib.importer.readData()`?)
# TODO: Exception subclasses for `get_doc()` failures, to separate the function's
#  own errors from `ValueError` exceptions raised by things the function calls?

from __future__ import annotations

import typing
from typing import Union, Tuple

from datetime import datetime
import os
from pathlib import Path
import tempfile
from urllib.parse import urlparse, ParseResult

from idelib.importer import openFile, readData
from idelib.util import extractTime
import requests

from .gdrive import gdrive_download
from .util import parse_time, validate

__all__ = ["get_doc", "extract_time"]

# ============================================================================
#
# ============================================================================


def normalized_path(
    filepath: str, mode: typing.Literal[None, "local", "url"] = None
) -> Union[
    Tuple[str, typing.Literal["local"]],
    Tuple[ParseResult, typing.Literal["url"]],
]:
    """
    Format a file path as either a URL, or a local file path.

    Local file paths are made into absolute paths via `os.path.abspath` and
    `os.path.expanduser`. URL paths are returned as a `urllib.parse.ParseResult`
    object.

    :param filepath: some file path
    :return: a tuple of 1) the formatted file path, and 2) the type of path
    """
    if mode not in (None, "local", "url"):
        raise ValueError(f"invalid mode {mode}; expected one of (None, 'local', 'url')")
    if mode is None:
        mode = "local" if os.path.isfile(filepath) else "url"

    if mode == "url":
        parsed_url = urlparse(filepath.replace("\\", "/"))
        if parsed_url.scheme == "file":
            filepath = parsed_url.path
            mode = "local"
        elif parsed_url.netloc:
            filepath = parsed_url
            mode = "url"
        else:
            mode = "local"

    if mode == "local":
        filepath = os.path.abspath(os.path.expanduser(filepath))

    return filepath, mode


# ============================================================================
#
# ============================================================================


def _get_url(url, localfile=None, headers=None, params=None, cookies=None):
    """
    Retrieve an IDE from a (HTTP/HTTPS) URL, including Google Drive shared
    links.

    :param url: The file's URL.
    :param localfile: The local filename (if saving the file).
    :param headers: Additional (optional) request headers.
    :param params: Additional (optional) request parameters.
    :param cookies: Optional browser cookies for the session.
    :return: An open file stream containing the IDE data and the number of
        bytes downloaded.
    """
    parsed_url = urlparse(url)
    session = requests.Session()

    netloc = parsed_url.netloc.lower()
    if netloc.endswith('.google.com') or netloc == "google.com":
        response, filename = gdrive_download(url, localfile, params=params,
                                             cookies=cookies)
    else:
        response = session.get(parsed_url.geturl(), headers=headers,
                               params=params, cookies=cookies)
        filename = None

    if not response.ok:
        raise ValueError(f"Could not retrieve data from URL {url} "
                         f"({response.status_code}: {response.reason})")

    if localfile is None:
        stream = tempfile.SpooledTemporaryFile(suffix=".ide")
    else:
        localfile = os.path.abspath(os.path.expanduser(localfile))
        if os.path.isdir(localfile):
            if not filename:
                filename = os.path.basename(parsed_url.path)
            localfile = os.path.join(localfile, filename)
        if not localfile.lower().endswith('.ide'):
            localfile += ".ide"
        stream = open(localfile, 'w+b')

    total = 0
    for chunk in response.iter_content(2**15):
        # future: Confirm that this is an IDE from 1st chunk, avoiding download if not
        if chunk:
            stream.write(chunk)
            total += len(chunk)

    response.close()
    stream.seek(0)

    return stream, total


# ============================================================================
#
# ============================================================================

def get_doc(name=None, filename=None, url=None, parsed=True, start=0, end=None,
            localfile=None, params=None, headers=None, cookies=None, **kwargs):
    """
    Retrieve an IDE file from either a file or URL.

    Note: `name`, `filename`, and `url` are mutually exclusive arguments.
    One and only one must be specified. Attempting to supply more than one
    will generate an error.

    Example usage::

        get_doc("my_recording.ide")
        get_doc("https://example.com/remote_recording.ide")
        get_doc(filename="my_recording.ide")
        get_doc(url="https://example.com/remote_recording.ide")
        get_doc(filename="my_recording.ide", start="1:23")

    The `start` and `end` times, if used, may be specified in several
    ways:

    * `int`/`float` (Microseconds from the recording start)
    * `str` (formatted as a time from the recording start, e.g., `MM:SS`,
      `HH:MM:SS`, `DDd HH:MM:SS`). More examples:

      * ``":01"`` or ``":1"`` or ``"1s"`` (1 second)
      * ``"22:11"`` (22 minutes, 11 seconds)
      * ``"3:22:11"`` (3 hours, 22 minutes, 11 seconds)
      * ``"1d 3:22:11"`` (1 day, 3 hours, 22 minutes, 11 seconds)
    * `datetime.timedelta` or `pandas.Timedelta` (time from the
      recording start)
    * `datetime.datetime` (an explicit UTC time)

    :param name: The name or URL of the IDE. The method of fetching it will
        be automatically chosen based on how it is formatted.
    :param filename: The name of an IDE file. Supplying a name this way will
        force it to be read from a file, avoiding the possibility of
        accidentally trying to retrieve it via URL.
    :param url: The URL of an IDE file. Supplying a name this way will force
        it to be read from a URL, avoiding the possibility of accidentally
        trying to retrieve it from a local file.
    :param parsed: If `True` (default), the IDE will be fully parsed after it
        is fetched. If `False`, only the file metadata will be initially
        loaded, and a call to `idelib.importer.readData()`. This can save
        time.
    :param start: The starting time. Defaults to the start of the
        recording. Only applicable if `parsed` is `True`.
    :param end: The ending time. Defaults to the end of the recording.  Only
        applicable if `parsed` is `True`.
    :param localfile: The name of the file to which to write data received
        from a URL. If none is supplied, a temporary file will be used. Only
        applicable when opening a URL.
    :param params: Additional URL request parameters. Only applicable when
        opening a URL.
    :param headers: Additional URL request headers. Only applicable when
        opening a URL.
    :param cookies: Additional browser cookies for use in the URL request.
        Only applicable when opening a URL.
    :return: The fetched IDE data.

    Additionally, `get_doc()` will accept the keyword arguments for
    `idelib.importer.importFile()` or `idelib.importer.openFile()`
    """
    if len([x for x in (name, filename, url) if x]) != 1:
        raise TypeError("Only one source can be specified: name, filename, or url")

    original = name or filename or url  # For error reporting
    stream = None

    if name:
        path_formatted, mode = normalized_path(name)
    elif url:
        kwargs.setdefault("name", url)
        path_formatted, mode = normalized_path(url, "url")
    elif filename:
        path_formatted, mode = normalized_path(filename, "local")
    # else not possible; one of name, url & filename is already guaranteed to be valid

    if mode == "url":
        parsed_url = path_formatted
        if parsed_url.scheme.startswith('http'):
            stream, _total = _get_url(parsed_url.geturl(), localfile=localfile, headers=headers,
                                      params=params, cookies=cookies)
        else:
            # future: more fetching schemes before this `else` (ftp, etc.)?
            raise ValueError(f"Unsupported transfer scheme: {parsed_url.scheme}")
    else:  # mode == "local":
        stream = open(path_formatted, 'rb')

    if stream:
        if not validate(stream):
            stream.close()
            raise ValueError(f"Could not read a Dataset from '{original}'"
                             f"(not an IDE file?)")

        # Separate `openFile()` and `readData` kwargs, remove ones that aren't shared
        open_kwargs = kwargs.copy()
        read_kwargs = kwargs.copy()

        for k in ('startTime', 'endTime', 'channels', 'source', 'total',
                  'bytesRead', 'samplesRead'):
            open_kwargs.pop(k, None)

        doc = openFile(stream, **open_kwargs)

        if parsed:
            for k in ('defaults', 'name', 'quiet'):
                read_kwargs.pop(k, None)

            session_start = doc.lastSession.utcStartTime
            if session_start:
                session_start = datetime.utcfromtimestamp(session_start)

            if start:
                read_kwargs['startTime'] = parse_time(start, session_start)
            if end:
                read_kwargs['endTime'] = parse_time(end, session_start)

            readData(doc, **read_kwargs)

        return doc

    raise ValueError(f"Could not read data from '{original}'")


def extract_time(doc, out, start=0, end=None, channels=None, **kwargs):
    """
    Efficiently extract data within a certain interval from an IDE file,
    writing it to another file. Note that due to the way data is stored in an
    IDE, the exported interval will be slightly wider than the specified
    start and end times; this ensures the data is copied verbatim and without
    loss.

    The `start` and `end` times, if used, may be specified in several
    ways:

    * `int`/`float` (Microseconds from the recording start)
    * `str` (formatted as a time from the recording start, e.g., `MM:SS`,
      `HH:MM:SS`, `DDd HH:MM:SS`). More examples:

      * ``":01"`` or ``":1"`` or ``"1s"`` (1 second)
      * ``"22:11"`` (22 minutes, 11 seconds)
      * ``"3:22:11"`` (3 hours, 22 minutes, 11 seconds)
      * ``"1d 3:22:11"`` (1 day, 3 hours, 22 minutes, 11 seconds)
    * `datetime.timedelta` or `pandas.Timedelta` (time from the
      recording start)
    * `datetime.datetime` (an explicit UTC time)

    :param doc: A `Dataset` or the name of a local IDE file. `Dataset`
        objects do not have to be fully imported.
    :param out: A filename or stream to which to save the extracted data.
    :param start: The starting time. Defaults to the start of the recording.
    :param end: The ending time. Defaults to the end of the recording.
    :param channels: A list of channel IDs to specifically export. If `None`,
        all channels will be exported. Note excluded channels will still
        appear in the new IDE's `channels` dictionary, but the file will
        contain no data for them.
    :return: The total number of bytes written, and total number of
        ChannelDataBlock elements copied.
    """
    if isinstance(doc, (str, Path)):
        doc = openFile(doc)

    session_start = doc.lastSession.utcStartTime
    if session_start:
        session_start = datetime.utcfromtimestamp(session_start)

    if start:
        kwargs['startTime'] = parse_time(start, session_start)
    if end:
        kwargs['endTime'] = parse_time(end, session_start)
    kwargs['channels'] = channels

    return extractTime(doc, out, **kwargs)



from typing import List, Tuple

import pandas as pd
import scipy.spatial


def _validate_euler_mode(mode: str) -> Tuple[str, List[str]]:

    _mode = mode.lower()

    names = (
        {'x', 'y', 'z'},
        {'roll', 'pitch', 'yaw'},
        {'alpha', 'beta', 'gamma'},
        {'α', 'β', 'γ'},
        {'psi', 'theta', 'phi'},
        {'ψ', 'θ', 'φ'},
        )

    separators = (
        ' ',
        '-',
        '_',
        )

    for s in separators:
        if s in _mode:
            mode_list = _mode.split(s)
            break
    else:
        if not set(_mode).issubset({'x', 'y', 'z'}):
            raise ValueError(f'Modes other than xyz (such as xyz, xyx, zxz) must '
                             f'separated with one of " ", "-" or "_".  Mode '
                             f'{mode} is not a valid euler angle mode.')
        mode_list = list(_mode)

    if not (
            # must have three rotations
            len(mode_list) == 3
            and (
                    # must have three unique rotations
                    len(set(mode_list)) == 3

                    # must have two unique rotations, and the first and last
                    # rotations must be about the same axes
                    or (len(set(mode_list)) == 2 and mode_list[0] == mode_list[2])
                )
            ):
        raise ValueError(f'Euler modes must have 3 elements '
                         f'and must have either three unique elements (Tait-Bryan angles) '
                         f'or must have two unique elements in the pattern x-y-x '
                         f'(proper Euler angles).  Mode {mode} does not fit this format')

    for n in names:
        if set(mode_list).issubset(n):
            new_mode = ''.join([a for m in mode_list for a, b in zip(names[0], n) if m == b])
            return new_mode, mode_list
    else:
        raise ValueError('Mode uses an unknown naming convention, valid naming '
                         'conventions are: x-y-z, roll-pitch-yaw, '
                         'alpha-beta-gamma, α-β-γ, psi-theta-phi, ψ-θ-φ')


def quaternion_to_euler(df: pd.DataFrame, mode: str = 'x-y-z') -> pd.DataFrame:
    """
    Convert quaternion data in the dataframe ``df`` to euler angles.  This can
    be done with either intrinsic or extrinsic rotations, determined
    automatically based on ``mode``.

    :param df:  The input quaternions to convert.  Must have columns labelled
                'X', 'Y', 'Z', and 'W'.
    :param mode: The order of the axes to rotate.  The default is intrinsic
                 rotation about x-y-z.
    :return:  A dataframe with the euler-angles of the quaternion data.

    .. seealso::
        - `SciPy's documentation on converting into euler angles <https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.transform.Rotation.as_euler.html>`_
        - `Wikipedia's article on Euler angles <https://en.wikipedia.org/wiki/Euler_angles>`_
    """

    valid_mode, columns = _validate_euler_mode(mode)

    r = scipy.spatial.transform.Rotation.from_quat(df[['X', 'Y', 'Z', 'W']])

    return pd.DataFrame(r.as_euler(valid_mode), index=df.index, columns=columns)



# -*- coding: utf-8 -*-

from __future__ import annotations

import typing
from typing import Optional, Union
import warnings

import numpy as np
import pandas as pd
import scipy.signal
import pint


def sample_spacing(
    data: Union[np.ndarray, pd.DataFrame],
    convert: typing.Literal[None, "to_seconds"] = "to_seconds",
) -> Union[None, float, np.timedelta64]:
    """
    Calculate the average spacing between individual samples.

    For time indices, this calculates the sampling period `dt`.

    :param data: the input data; either a pandas DataFrame with the samples
        spaced along its index, or a 1D-array-like of sample times
    :param convert: if `"to_seconds"` (default), convert any time objects into
        floating-point seconds
    """
    if isinstance(data, (pd.DataFrame, pd.Series)):
        data = data.index
    if len(data) <= 1:
        return None

    dt = (data[-1] - data[0]) / (len(data) - 1)
    if convert == "to_seconds" and isinstance(dt, (np.timedelta64, pd.Timedelta)):
        dt = dt / np.timedelta64(1, "s")

    return dt


def logfreqs(
    df: pd.DataFrame, init_freq: Optional[float] = None, bins_per_octave: float = 12.0
) -> np.ndarray:
    """
    Calculate a sequence of log-spaced frequencies for a given dataframe.

    :param df: the input data
    :param init_freq: the initial frequency in the sequence; if `None` (default),
        use the frequency corresponding to the data's duration
    :param bins_per_octave: the number of frequencies per octave
    :return: an array of log-spaced frequencies
    """
    dt = sample_spacing(df)
    T = dt * len(df.index)

    if init_freq is None:
        init_freq = 1 / T
    elif 1 / init_freq > T:
        warnings.warn(
            "the data's duration is too short to accurately represent an"
            f" initial frequency of {init_freq:.3f} Hz",
            RuntimeWarning,
        )

    return 2 ** np.arange(
        np.log2(init_freq),
        np.log2(1 / dt) - 1,
        1 / bins_per_octave,
    )


def to_dB(
    data: np.ndarray,
    reference: Union[float, typing.Literal[tuple(dB_refs.keys())]],
    squared: bool = False,
) -> np.ndarray:
    """
    Scale data into units of decibels.

    Decibels are a log-scaled ratio of some value against a reference;
    typically this is expressed as follows:

    .. math:: dB = 10 \\log_{10}\\left( \\frac{x}{x_{\\text{ref}}} \\right)

    By convention, "decibel" units tend to operate on units of *power*. For
    units that are proportional to power *when squared* (e.g., volts, amps,
    pressure, etc.), their "decibel" representation is typically doubled (i.e.,
    :math:`dB = 20 \\log_{10}(...)`). Users can specify which scaling to use
    with the `squared` parameter.

    .. note::
        Decibels can **NOT** be calculated from negative values.

        For example, to calculate dB on arbitrary time-series data, typically
        data is first aggregated via:

        - a *total RMS* (like :py:func:`endaq.calc.stats.rms`),
        - a *rolling RMS* (like :py:func:`endaq.calc.stats.rolling_rms`), or
        - a *PSD* (like :py:func:`endaq.calc.psd.welch`),

        and the non-negative result from the aggregation is then scaled into
        decibels.

    :param data: the input data
    :param reference: the reference value corresponding to 0dB
    :param squared: whether the input data & reference value are pre-squared;
        defaults to `False`
    """
    if isinstance(reference, str):
        try:
            reference = dB_refs[reference]
        except KeyError:
            raise ValueError(f'unknown reference "{reference}"')
    elif reference <= 0:
        raise ValueError("reference value must be strictly positive")

    data = np.asarray(data)
    if np.any(data < 0):
        raise ValueError(
            "cannot compute decibels from negative values (see the docstring"
            " for details)"
        )

    return (10 if squared else 20) * (np.log10(data) - np.log10(reference))


dB_refs = {
    "SPL": 2e-5,  # Pascal
    "audio_intensity": 1e-12,  # W/m²
}


def resample(df: pd.DataFrame, sample_rate: Optional[float] = None) -> pd.DataFrame:
    """
    Resample a dataframe to a desired sample rate (in Hz)

    :param df: The DataFrame to resample, indexed by time
    :param sample_rate: The desired sample rate to resample the given data to.
     If one is not supplied, then it will use the same as it currently does, but
     make the time stamps uniformly spaced
    :return: The resampled data in a DataFrame
    """
    if sample_rate is None:
        num_samples_after_resampling = len(df)
    else:
        dt = sample_spacing(df)
        num_samples_after_resampling = int(dt * len(df) * sample_rate)

    resampled_data, resampled_time = scipy.signal.resample(
        df,
        num_samples_after_resampling,
        t=df.index.values.astype(np.float64),
    )

    # Check for datetimes, if so localize
    if 'datetime' in str(df.index.dtype):
        df.index = df.index.tz_localize(None)

    resampled_df = pd.DataFrame(
        resampled_data,
        index=resampled_time.astype(df.index.dtype),
        columns=df.columns,
    )
    
    resampled_df.index.name = df.index.name
    
    return resampled_df


def _rolling_slice_definitions(
        df: pd.DataFrame,
        num_slices: int = 100,
        indexes: np.array = None,
        index_values: np.array = None,
        slice_width: float = None,
):
    """
    Compute parameters needed to define index locations and slice width for rolling computations

    :param df: the input dataframe with an index defining the time in seconds or datetime
    :param num_slices: the number of slices to split the time series into, default is 100,
        this is ignored if `indexes` is defined
    :param indexes: the center index locations (not value) of each slice to compute the FFT
    :param index_values: the index values of each peak event to quantify (slower but more intuitive than using `indexes`)
    :param slice_width: the time in seconds to center about each slice index,
        if none is provided it will calculate one based upon the number of slices
    :return: a tuple of `indexes`, `slice_width`, `num`, and `length`

    See example use cases and syntax at :py:func:`~endaq.plot.spectrum_over_time()`
    which visualizes the output of this function in Heatmaps, Waterfall plots,
    Surface plots, and Animations

    """

    length = len(df)

    # Define center index locations of each slice if not provided
    if indexes is None:
        if index_values is not None:
            indexes = np.zeros(len(index_values), int)
            for i in range(len(indexes)):
                indexes[i] = int((np.abs(df.index - index_values[i])).argmin())
        else:
            indexes = np.linspace(0, length, num_slices, endpoint=False, dtype=int)
            indexes = indexes + int(indexes[1] / 2)

    # Calculate slice step size
    spacing = sample_spacing(df)
    if slice_width is None:
        slice_width = spacing * length / len(indexes)
    num = int(slice_width / spacing / 2)

    return indexes, slice_width, num, length


def convert_units(
        units_in: str,
        units_out: str,
        df: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """
    Using the `Pint library <https://github.com/hgrecco/pint/blob/master/pint/default_en.txt>`_ apply a unit
    conversion to a provided unit-unaware dataframe.

    :param units_in: a text string defining the base units to convert from like `"in"` for inches
    :param units_out: a text string defining the destination units to convert to like `"mm"` for millimeters
    :param df: the input dataframe, if none the unit conversion is only applied from `units_in` to `units_out`
    :return: a dataframe with the values scaled according to the unit conversion, if no dataframe is provided then a
        scaler value is returned

    Some examples are provided below which includes a table of common unit conversions. A full list is available from
    the `Pint library <https://github.com/hgrecco/pint/blob/master/pint/default_en.txt>`_.

    .. code:: python3

        import endaq
        endaq.plot.utilities.set_theme()
        import pandas as pd
        import plotly.express as px

        # Simple conversion factor from inches to millimeters
        in_2_mm = endaq.calc.utils.convert_units('in', 'mm')

        # Get idelib dataset
        doc = endaq.ide.get_doc('https://info.endaq.com/hubfs/data/All-Channels.ide')

        # Get acceleration data in 'g' and convert to in/s^2
        accel_in_gs = endaq.ide.get_primary_sensor_data(doc=doc, measurement_type='accel')
        accel_in_inches = endaq.calc.utils.convert_units('gravity', 'in/s**2', accel_in_gs)

        # Get temperature in Celsius and convert to Fahrenheit
        temp_in_C = endaq.ide.get_primary_sensor_data(doc=doc, measurement_type='temp')
        temp_in_F = endaq.calc.utils.convert_units('degC', 'degF', temp_in_C)

        # Merge C & F in one dataframe
        temp_in_C.columns = ['Temperature in Degrees C']
        temp_in_F.columns = ['Temperature in Degrees F']
        temp = pd.concat([temp_in_C, temp_in_F], axis=1)

        # Display plot with both C and F
        fig = px.line(
            temp.reset_index().melt(id_vars='timestamp'),
            x='timestamp',
            y='value',
            facet_col='variable',
            facet_col_spacing=0.07
        ).for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
        fig.update_yaxes(matches=None, showticklabels=True, title_text='').update_xaxes(title_text='')
        fig.show()

        # Get Table of Unit Conversions
        df = pd.read_csv('https://info.endaq.com/hubfs/Unit-Conversion-Examples.csv')
        df['output'] = 0
        for i in df.index:
            df.loc[i, 'output'] = endaq.calc.utils.convert_units(
                units_in = df.loc[i, 'units_in'],
                units_out = df.loc[i, 'units_out'])

        # Generate Plot Table
        plot_table = endaq.plot.table_plot(df, num_round=6)
        plot_table.show()

    .. plotly::
        :fig-vars: fig, plot_table

        import endaq
        endaq.plot.utilities.set_theme()
        import pandas as pd
        import plotly.express as px

        # Simple conversion factor from inches to millimeters
        in_2_mm = endaq.calc.utils.convert_units('in', 'mm')

        # Get idelib dataset
        doc = endaq.ide.get_doc('https://info.endaq.com/hubfs/data/All-Channels.ide')

        # Get acceleration data in 'g' and convert to in/s^2
        accel_in_gs = endaq.ide.get_primary_sensor_data(doc=doc, measurement_type='accel')
        accel_in_inches = endaq.calc.utils.convert_units('gravity', 'in/s**2', accel_in_gs)

        # Get temperature in Celsius and convert to Fahrenheit
        temp_in_C = endaq.ide.get_primary_sensor_data(doc=doc, measurement_type='temp')
        temp_in_F = endaq.calc.utils.convert_units('degC', 'degF', temp_in_C)

        # Merge C & F in one dataframe
        temp_in_C.columns = ['Temperature in Degrees C']
        temp_in_F.columns = ['Temperature in Degrees F']
        temp = pd.concat([temp_in_C, temp_in_F], axis=1)

        # Display plot with both C and F
        fig = px.line(
            temp.reset_index().melt(id_vars='timestamp'),
            x='timestamp',
            y='value',
            facet_col='variable',
            facet_col_spacing=0.07
        ).for_each_annotation(lambda a: a.update(text=a.text.split("=")[-1]))
        fig.update_yaxes(matches=None, showticklabels=True, title_text='').update_xaxes(title_text='')
        fig.show()

        # Get Table of Unit Conversions
        df = pd.read_csv('https://info.endaq.com/hubfs/Unit-Conversion-Examples.csv')
        df['output'] = 0
        for i in df.index:
            df.loc[i, 'output'] = endaq.calc.utils.convert_units(
                units_in = df.loc[i, 'units_in'],
                units_out = df.loc[i, 'units_out'])

        # Generate Plot Table
        plot_table = endaq.plot.table_plot(df, num_round=6)
        plot_table.show()

    """
    ureg = pint.UnitRegistry(autoconvert_offset_to_baseunit = True)
    src = ureg(units_in)
    dst = ureg(units_out)

    if df is None:
        return src.to(dst).magnitude
    else:
        converted_df = df.copy()
        vals = (df.values * src).to(dst).magnitude
        for i, c in enumerate(df.columns):
            converted_df[c] = vals[:, i]
        return converted_df



# -*- coding: utf-8 -*-

from __future__ import annotations

import typing
from typing import Optional
from collections import namedtuple
from dataclasses import dataclass
import functools
import warnings

import numpy as np
import pandas as pd
import scipy.signal

from endaq.calc import utils, stats


def _absolute_acceleration_coefficients(omega, Q, T):
    """
    Calculate the coefficients of the Z-domain transfer function for the
    absolute acceleration response according to ISO 18431-4.

    :param omega: the natural frequency of the system
    :param Q: the quality factor of the system
    :param T: the time step in seconds
    :return: the coefficients of the Z-domain transfer function b, a

    .. seealso::

        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    A = omega*T/(2.*Q)
    B = omega*T*np.sqrt(1. - 1./(4.*(Q**2)))

    b = (
        1. - np.exp(-A)*np.sin(B)/B,
        2.*np.exp(-A)*(np.sin(B)/B - np.cos(B)),
        np.exp(-2*A) - np.exp(-A)*np.sin(B)/B,
        )
    a = (
        1.,
        -2.*np.exp(-A)*np.cos(B),
        np.exp(-2.*A),
        )

    return b, a


def absolute_acceleration(accel: pd.DataFrame, omega: float, damp: float = 0.0) -> pd.DataFrame:
    """
    Calculate the absolute acceleration for a SDOF system.

    The absolute acceleration follows the transfer function:

        `H(s) = L{x"(t)}(s) / L{y"(t)}(s) = X(s)/Y(s)`

    for the PDE:

        `x" + (2ζω)x' + (ω²)x = (2ζω)y' + (ω²)y`

    :param accel: the absolute acceleration `y"`
    :param omega: the natural frequency `ω` of the SDOF system
    :param damp: the damping coefficient `ζ` of the SDOF system
    :return: the absolute acceleration `x"` of the SDOF system

    .. seealso::

        - `An Introduction To The Shock Response Spectrum, Tom Irvine, 9 July 2012 <http://www.vibrationdata.com/tutorials2/srs_intr.pdf>`_
        - `SciPy transfer functions <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.TransferFunction.html>`_
          Documentation for the transfer function class used to characterize the
          relative displacement calculation.
        - `SciPy biquad filter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.lfilter.html>`__
          Documentation for the biquad function used to implement the transfer
          function.
        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    T = utils.sample_spacing(accel)
    Q = 1./(2.*damp)

    return accel.apply(
        functools.partial(
                scipy.signal.lfilter,
                *_absolute_acceleration_coefficients(omega, Q, T),
                axis=0,
                ),
        raw=True,
        )


def _relative_velocity_coefficients(omega, Q, T):
    """
    Calculate the coefficients of the Z-domain transfer function for the
    relative velocity response according to ISO 18431-4.

    :param omega: the natural frequency of the system
    :param Q: the quality factor of the system
    :param T: the time step in seconds
    :return: the coefficients of the Z-domain transfer function b, a

    .. seealso::

        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    A = omega*T/(2.*Q)
    B = omega*T*np.sqrt(1. - 1./(4.*(Q**2.)))
    C = np.exp(-A)*np.sin(B)/np.sqrt(4.*(Q**2.) - 1.)
    D = T*(omega**2.)

    b = (
        (-1. + np.exp(-A)*np.cos(B) + C)/D,
        (1. - np.exp(-2.*A) - 2.*C)/D,
        (np.exp(-2.*A) - np.exp(-A)*np.cos(B) + C)/D,
        )
    a = (
        1.,
        -2.*np.exp(-A)*np.cos(B),
        np.exp(-2.*A),
        )

    return b, a


def relative_velocity(accel: pd.DataFrame, omega: float, damp: float = 0.0) -> pd.DataFrame:
    """
    Calculate the relative velocity for a SDOF system.

    The relative velocity follows the transfer function:

        `H(s) = L{z'(t)}(s) / L{y"(t)}(s) = (1/s)(Z(s)/Y(s))`

    for the PDE:

        `z" + (2ζω)z' + (ω²)z = -y"`

    :param accel: the absolute acceleration y"
    :param omega: the natural frequency ω of the SDOF system
    :param damp: the damping coefficient ζ of the SDOF system
    :return: the relative velocity z' of the SDOF system

    .. seealso::

        - `Pseudo Velocity Shock Spectrum Rules For Analysis Of Mechanical Shock, Howard A. Gaberson <https://info.endaq.com/hubfs/pvsrs_rules.pdf>`_
        - `SciPy transfer functions <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.TransferFunction.html>`_
          Documentation for the transfer function class used to characterize the
          relative displacement calculation.
        - `SciPy biquad filter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.lfilter.html>`__
          Documentation for the biquad function used to implement the transfer
          function.
        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    T = utils.sample_spacing(accel)
    Q = 1./(2.*damp)

    return accel.apply(
        functools.partial(
                scipy.signal.lfilter,
                *_relative_velocity_coefficients(omega, Q, T),
                axis=0,
                ),
        raw=True,
        )


def _relative_displacement_coefficients(omega, Q, T):
    """
    Calculate the coefficients of the Z-domain transfer function for the
    relative displacement response according to ISO 18431-4.

    :param omega: the natural frequency of the system
    :param Q: the quality factor of the system
    :param T: the time step in seconds
    :return: the coefficients of the Z-domain transfer function b, a

    .. seealso::

        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    A = omega*T/(2.*Q)
    B = omega*T*np.sqrt(1. - 1./(4.*(Q**2.)))
    C = T*(omega**3.)
    q = (1./(2.*(Q**2.)) - 1.)/np.sqrt(1. - 1./(4.*(Q**2.)))

    b = (
        ((1. - np.exp(-A)*np.cos(B))/Q - q*np.exp(-A)*np.sin(B) - omega*T)/C,
        (2.*np.exp(-A)*np.cos(B)*omega*T -
         (1. - np.exp(-2.*A))/Q +
         2*q*np.exp(-A)*np.sin(B))/C,
        (-np.exp(-2.*A)*(omega*T + 1./Q) +
         np.exp(-A)*np.cos(B)/Q -
         q*np.exp(-A)*np.sin(B))/C,
        )
    a = (
        1.,
        -2.*np.exp(-A)*np.cos(B),
        np.exp(-2.*A),
        )

    return b, a


def relative_displacement(accel: pd.DataFrame, omega: float, damp: float = 0.0) -> pd.DataFrame:
    """
    Calculate the relative displacement for a SDOF system.

    The relative displacement follows the transfer function:

        `H(s) = L{z(t)}(s) / L{y"(t)}(s) = (1/s²)(Z(s)/Y(s))`

    for the PDE:

        `z" + (2ζω)z' + (ω²)z = -y"`

    :param accel: the absolute acceleration y"
    :param omega: the natural frequency ω of the SDOF system
    :param damp: the damping coefficient ζ of the SDOF system
    :return: the relative displacement z of the SDOF system

    .. seealso::

        - `Pseudo Velocity Shock Spectrum Rules For Analysis Of Mechanical Shock, Howard A. Gaberson <https://info.endaq.com/hubfs/pvsrs_rules.pdf>`_
        - `SciPy transfer functions <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.TransferFunction.html>`_
          Documentation for the transfer function class used to characterize the
          relative displacement calculation.
        - `SciPy biquad filter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.lfilter.html>`__
          Documentation for the biquad function used to implement the transfer
          function.
        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    T = utils.sample_spacing(accel)
    Q = 1./(2.*damp)

    return accel.apply(
        functools.partial(
                scipy.signal.lfilter,
                *_relative_displacement_coefficients(omega, Q, T),
                axis=0,
                ),
        raw=True,
        )


def _pseudo_velocity_coefficients(omega, Q, T):
    """
    Calculate the coefficients of the Z-domain transfer function for the
    pseudo-velocity response according to ISO 18431-4.

    :param omega: the natural frequency of the system
    :param Q: the quality factor of the system
    :param T: the time step in seconds
    :return: the coefficients of the Z-domain transfer function b, a

    .. seealso::

        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    A = omega*T/(2.*Q)
    B = omega*T*np.sqrt(1. - 1./(4.*(Q**2)))
    C = T*(omega**2)
    q = (1./(2.*(Q**2.)) - 1.)/np.sqrt(1. - 1./(4.*(Q**2.)))

    b = (
        ((1. - np.exp(-A)*np.cos(B))/Q - q*np.exp(-A)*np.sin(B) - omega*T)/C,
        (2.*np.exp(-A)*np.cos(B)*omega*T - (1. - np.exp(-2.*A))/Q + 2.*q*np.exp(-A)*np.sin(B))/C,
        (-np.exp(-2.*A)*(omega*T + 1./Q) + np.exp(-A)*np.cos(B)/Q - q*np.exp(-A)*np.sin(B))/C,
        )
    a = (
        1.,
        -2.*np.exp(-A)*np.cos(B),
        np.exp(-2.*A),
        )

    return b, a


def pseudo_velocity(accel: pd.DataFrame, omega: float, damp: float = 0.0) -> pd.DataFrame:
    """
    Calculate the pseudo-velocity for a SDOF system.

    The pseudo-velocity follows the transfer function:

        `H(s) = L{ωz(t)}(s) / L{y"(t)}(s) = (ω/s²)(Z(s)/Y(s))`

    for the PDE:

        `z" + (2ζω)z' + (ω²)z = -y"`

    :param accel: the absolute acceleration y"
    :param omega: the natural frequency ω of the SDOF system
    :param damp: the damping coefficient ζ of the SDOF system
    :return: the pseudo-velocity of the SDOF system

    .. seealso::

        - `Pseudo Velocity Shock Spectrum Rules For Analysis Of Mechanical Shock, Howard A. Gaberson <https://info.endaq.com/hubfs/pvsrs_rules.pdf>`_
        - `SciPy transfer functions <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.TransferFunction.html>`_
          Documentation for the transfer function class used to characterize the
          relative displacement calculation.
        - `SciPy biquad filter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.lfilter.html>`__
          Documentation for the biquad function used to implement the transfer
          function.
        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    T = utils.sample_spacing(accel)
    Q = 1./(2.*damp)

    return accel.apply(
        functools.partial(
                scipy.signal.lfilter,
                *_pseudo_velocity_coefficients(omega, Q, T),
                axis=0,
                ),
        raw=True,
        )


def _relative_displacement_static_coefficients(omega, Q, T):
    """
    Calculate the coefficients of the Z-domain transfer function for the
    relative displacement response expressed as equivalent static acceleration
    according to ISO 18431-4.

    :param omega: the natural frequency of the system
    :param Q: the quality factor of the system
    :param T: the time step in seconds
    :return: the coefficients of the Z-domain transfer function b, a

    .. seealso::

        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    A = omega*T/(2.*Q)
    B = omega*T*np.sqrt(1. - 1/(4.*(Q**2.)))
    C = (T*omega)
    q = (1./(2.*(Q**2.)) - 1.)/(np.sqrt(1. - 1./(4.*(Q**2.))))

    b = (
        ((1 - np.exp(-A)*np.cos(B))/Q - q*np.exp(-A)*np.sin(B) - omega*T)/C,
        (2*np.exp(-A)*np.cos(B)*omega*T - (1 - np.exp(-2.*A))/Q + 2.*q*np.exp(-A)*np.sin(B))/C,
        (-np.exp(-2.*A)*(omega*T + 1./Q) + np.exp(-A)*np.cos(B)/Q - q*np.exp(-A)*np.sin(B))/C,
        )
    a = (
        1.,
        -2.*np.exp(-A)*np.cos(B),
        np.exp(-2.*A),
        )

    return b, a


def relative_displacement_static(accel: pd.DataFrame, omega: float, damp: float = 0.0) -> pd.DataFrame:
    """
    Calculate the relative displacement expressed as equivalent static
    acceleration for a SDOF system.

    The relative displacement as static acceleration follows the transfer
    function:

        `H(s) = L{ω²z(t)}(s) / L{y"(t)}(s) = (ω²/s²)(Z(s)/Y(s))`

    for the PDE:

        `z" + (2ζω)z' + (ω²)z = -y"`

    :param accel: the absolute acceleration y"
    :param omega: the natural frequency ω of the SDOF system
    :param damp: the damping coefficient ζ of the SDOF system
    :return: the relative displacement of the SDOF system expressed as
        equivalent static acceleration

    .. seealso::

        - `Pseudo Velocity Shock Spectrum Rules For Analysis Of Mechanical Shock, Howard A. Gaberson <https://info.endaq.com/hubfs/pvsrs_rules.pdf>`_
        - `SciPy transfer functions <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.TransferFunction.html>`_
          Documentation for the transfer function class used to characterize the
          relative displacement calculation.
        - `SciPy biquad filter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.lfilter.html>`__
          Documentation for the biquad function used to implement the transfer
          function.
        - `ISO 18431-4 Mechanical vibration and shock — Signal processing — Part 4: Shock-response spectrum analysis`
          Explicit implementations of digital filter coefficients for shock spectra.
    """
    T = utils.sample_spacing(accel)
    Q = 1./(2.*damp)

    return accel.apply(
        functools.partial(
                scipy.signal.lfilter,
                *_relative_displacement_static_coefficients(omega, Q, T),
                axis=0,
                ),
        raw=True,
        )


def shock_spectrum(
    accel: pd.DataFrame,
    freqs: np.ndarray = None,
    init_freq: float = 0.5,
    bins_per_octave: float = 12.0,
    damp: float = 0.05,
    mode: typing.Literal["srs", "pvss"] = "srs",
    max_time: float = 2.0,
    peak_threshold: float = 0.1,
    two_sided: bool = False,
    aggregate_axes: bool = False,
) -> pd.DataFrame:
    """
    Calculate the shock spectrum of an acceleration signal. Note this defaults to first find peak events, then compute
    the spectrum on those peaks to speed up processing time.
    
    :param accel: the absolute acceleration `y"`
    :param freqs: the natural frequencies across which to calculate the spectrum,
        if `None` (the default) it uses `init_freq` and `bins_per_octave` to define them
    :param init_freq: the initial frequency in the sequence; if `None`,
        use the frequency corresponding to the data's duration, default is 0.5 Hz
    :param bins_per_octave: the number of frequencies per octave, default is 12    
    :param damp: the damping coefficient `ζ`, related to the Q-factor by
        `ζ = 1/(2Q)`; defaults to 0.05
    :param mode: the type of spectrum to calculate:

        *  `'srs'` (default) specifies the Shock Response Spectrum (SRS)
        *  `'pvss'` specifies the Pseudo-Velocity Shock Spectrum (PVSS)
    :param max_time: the maximum duration in seconds to compute the shock spectrum for, if the time duration is greater
        than :py:func:`~endaq.calc.stats.find_peaks()` is used to find peak locations, then the shock spectrums at each
        peak is calculated with :py:func:`~endaq.calc.shock.rolling_shock_spectrum()` with `max_time` defining the
        `slice_width`. Set `max_time` to `None` to force the function to not do the peak finding.
    :param peak_threshold: if the duration is greater than `max_time` all peaks that are greater than `peak_threshold`
        will be calculated, and the aggregate max per frequency will be reported
    :param two_sided: whether to return for each frequency:
        both the maximum negative and positive shocks (`True`),
        or simply the maximum absolute shock (`False`; default)
    :param aggregate_axes: whether to calculate the column-wise resultant (`True`)
        or calculate spectra along each column independently (`False`; default)
    :return: the shock spectrum
    .. seealso::
        - `Pseudo Velocity Shock Spectrum Rules For Analysis Of Mechanical Shock, Howard A. Gaberson <https://info.endaq.com/hubfs/pvsrs_rules.pdf>`_
        - `An Introduction To The Shock Response Spectrum, Tom Irvine, 9 July 2012 <http://www.vibrationdata.com/tutorials2/srs_intr.pdf>`_
        - `SciPy transfer functions <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.TransferFunction.html>`_
          Documentation for the transfer function class used to characterize the
          relative displacement calculation.
        - `SciPy biquad filter <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.lfilter.html>`__
          Documentation for the biquad function used to implement the transfer
          function.
    """
    if freqs is None:
        freqs = utils.logfreqs(accel, init_freq=init_freq, bins_per_octave=bins_per_octave)
    if two_sided and aggregate_axes:
        raise ValueError("cannot enable both options `two_sided` and `aggregate_axes`")
    freqs = np.asarray(freqs)
    if freqs.ndim != 1:
        raise ValueError("target frequencies must be in a 1D-array")

    # Check for total time, if too long compute only for peaks
    dt = utils.sample_spacing(accel)
    if max_time is None:
        max_time = dt * len(accel) * 2
    if dt * len(accel) > max_time:
        rolling_srs = rolling_shock_spectrum(
            accel,
            damp=damp,
            mode=mode,
            add_resultant=aggregate_axes,
            freqs=freqs,
            indexes=stats.find_peaks(accel, time_distance=max_time, threshold_multiplier=peak_threshold),
            slice_width=max_time
        )
        var_name = 'variable'
        if 'axis' in rolling_srs.columns:
            var_name = 'axis'
        srs = pd.DataFrame(
            index=pd.Series(rolling_srs['frequency (Hz)'].unique(), name='frequency (Hz)'),
            columns=rolling_srs[var_name].unique()
        )
        for var in srs.columns:
            pivot = rolling_srs[rolling_srs[var_name] == var].copy()
            pivot = pivot.pivot_table(
                columns='timestamp',
                index='frequency (Hz)',
                values='value'
            )
            srs[var] = pivot.max(axis=1)
        srs.columns.name = accel.columns.name
        return srs

    omega = 2 * np.pi * freqs

    if mode == "srs":
        make_coeffs = _absolute_acceleration_coefficients
    elif mode == "pvss":
        make_coeffs = _pseudo_velocity_coefficients
    else:
        raise ValueError(f"invalid spectrum mode {mode:r}")

    results = np.empty(
        (2,) + freqs.shape + ((1,) if aggregate_axes else accel.shape[1:]),
        dtype=np.float64,
    )

    T_padding = 1 / (
        freqs.min() * np.sqrt(1 - damp ** 2)
    )  # uses lowest damped frequency
    if not two_sided:
        T_padding /= 2

    zi = np.zeros((2,) + accel.shape[1:])
    zero_padding = np.zeros((int(T_padding // dt) + 1,) + accel.shape[1:])

    Q = 1./(2.*damp)

    for i_nd in np.ndindex(freqs.shape[0]):
        rd, zf = scipy.signal.lfilter(
                *make_coeffs(omega[i_nd], Q, dt),
                accel.to_numpy(),
                zi=zi,
                axis=0,
                )
        rd_padding, _ = scipy.signal.lfilter(
            *make_coeffs(omega[i_nd], Q, dt), zero_padding, zi=zf, axis=0
        )

        if aggregate_axes:
            rd = stats.L2_norm(rd, axis=-1, keepdims=True)
            rd_padding = stats.L2_norm(rd_padding, axis=-1, keepdims=True)

        results[(0,) + i_nd] = -np.minimum(rd.min(axis=0), rd_padding.min(axis=0))
        results[(1,) + i_nd] = np.maximum(rd.max(axis=0), rd_padding.max(axis=0))

    if aggregate_axes or not two_sided:
        return pd.DataFrame(
            np.maximum(results[0], results[1]),
            index=pd.Series(freqs, name="frequency (Hz)"),
            columns=(["Resultant"] if aggregate_axes else accel.columns),
        )

    return namedtuple("PseudoVelocityResults", "neg pos")(
        *(
            pd.DataFrame(
                r, index=pd.Series(freqs, name="frequency (Hz)"), columns=accel.columns
            )
            for r in results
        )
    )


def rolling_shock_spectrum(
        df: pd.DataFrame,
        damp: float = 0.05,
        mode: typing.Literal["srs", "pvss"] = "srs",
        add_resultant: bool = True,
        freqs: np.ndarray = None,
        init_freq: float = 0.5,
        bins_per_octave: float = 12.0,
        num_slices: int = 100,
        indexes: np.array = None,
        index_values: np.array = None,
        slice_width: float = None,
        disable_warnings: bool = True,
) -> pd.DataFrame:
    """
    Compute Shock Response Spectrums for defined slices of a time series data set using :py:func:`~endaq.calc.shock.shock_spectrum()`
    
    :param df: the input dataframe with an index defining the time in seconds or datetime
    :param damp: the damping coefficient `ζ`, related to the Q-factor by
        `ζ = 1/(2Q)`; defaults to 0.05
    :param mode: the type of spectrum to calculate:

        *  `'srs'` (default) specifies the Shock Response Spectrum (SRS)
        *  `'pvss'` specifies the Pseudo-Velocity Shock Spectrum (PVSS)
    :param add_resultant: if `True` (default) the column-wise resultant will
        also be computed and returned with the spectra per-column
    :param freqs: the natural frequencies across which to calculate the spectrum,
        if `None` (the default) it uses `init_freq` and `bins_per_octave` to define them
    :param init_freq: the initial frequency in the sequence; if `None`,
        use the frequency corresponding to the data's duration, default is 0.5 Hz
    :param bins_per_octave: the number of frequencies per octave, default is 12    
    :param num_slices: the number of slices to split the time series into, default is 100,
        this is ignored if `indexes` is defined
    :param indexes: the center index locations (not value) of each slice to compute the shock spectrum
    :param index_values: the index values of each peak event to quantify (slower but more intuitive than using `indexes`)
    :param slice_width: the time in seconds to center about each slice index,
        if none is provided it will calculate one based upon the number of slices
    :param disable_warnings: if `True` (default) it disables the warnings on the initial frequency
    :return: a dataframe containing all the shock spectrums, stacked on each other
    
    See example use cases and syntax at :py:func:`~endaq.plot.spectrum_over_time()`
    which visualizes the output of this function in Heatmaps, Waterfall plots, 
    Surface plots, and Animations

    """
    if freqs is None:
        freqs = utils.logfreqs(df, init_freq=init_freq, bins_per_octave=bins_per_octave)

    indexes, slice_width, num, length = utils._rolling_slice_definitions(
        df,
        indexes=indexes,
        index_values=index_values,
        num_slices=num_slices,
        slice_width=slice_width
    )

    # Loop through and compute shock spectrum
    srs = pd.DataFrame()
    for i in indexes:
        window_start = max(0, i - num)
        window_end = min(length, i + num)
        with warnings.catch_warnings():
            if disable_warnings:
                warnings.filterwarnings('ignore', '.*too short*', )
            slice_srs = shock_spectrum(
                df.iloc[window_start:window_end],
                mode=mode,
                damp=damp,
                freqs=freqs,
            )
        if add_resultant:
            with warnings.catch_warnings():
                if disable_warnings:
                    warnings.filterwarnings('ignore', '.*too short*', )
                slice_srs['Resultant'] = shock_spectrum(
                    df.iloc[window_start:window_end],
                    mode=mode,
                    damp=damp,
                    freqs=freqs,
                    aggregate_axes=True
                )['Resultant']

        slice_srs = slice_srs.reset_index().melt(id_vars=slice_srs.index.name)
        slice_srs['timestamp'] = df.index[i]
        srs = pd.concat([srs, slice_srs])

    return srs


@dataclass
class HalfSineWavePulse:
    """
    The output data type for :py:func:`enveloping_half_sine`.

    The significant data members are `amplitude` and `duration`, which can
    simply be unpacked as if from a plain tuple:

    .. testsetup::

        import pandas as pd
        df_pvss = pd.DataFrame([1, 1], index=[200, 400])

        from endaq.calc.shock import enveloping_half_sine

    .. testcode::

        ampl, T = enveloping_half_sine(df_pvss)

    However, users can also elect to use the other methods of this class to
    generate other kinds of outputs.

    .. note:: This class is not intended to be instantiated manually.
    """

    amplitude: pd.Series
    duration: pd.Series

    def __iter__(self):
        return iter((self.amplitude, self.duration))

    def to_time_series(
        self,
        tstart: Optional[float] = None,
        tstop: Optional[float] = None,
        dt: Optional[float] = None,
        tpulse: Optional[float] = None,
    ) -> pd.DataFrame:
        """
        Generate a time-series of the half-sine pulse.

        :param tstart: the starting time of the resulting waveform; if `None`
            (default), the range starts at `tpulse`
        :param tstop: the ending time of the resulting waveform; if `None`
            (default), the range ends at `tpulse + duration`
        :param dt: the sampling period of the resulting waveform; defaults to
            1/20th of the pulse duration
        :param tpulse: the starting time of the pulse within the resulting
            waveform; if `None` (default), the pulse starts at either:

            - ``tstart``, if provided
            - ``tstop - self.duration.max())``, if `tstop` is provided
            - ``0.0`` otherwise
        :return: a time-series of the half-sine pulse
        """
        if dt is None:
            dt = self.duration.min() / 20
        if dt > self.duration.min() / 8:
            warnings.warn(
                f"the sampling period {dt} is large relative to the pulse duration"
                f" {self.duration.min()}; the waveform may not accurately represent"
                f" the half-sine pulse's shock intensity"
            )

        default_start = 0.0
        if tstop is not None:
            default_start = tstop - self.duration.max()

        if tpulse is None and tstart is None:
            tpulse = tstart = default_start
        elif tpulse is None:
            tpulse = tstart
        elif tstart is None:
            tstart = tpulse

        if tstop is None:
            tstop = tpulse + self.duration.max()

        if not (tstart <= tpulse <= tstop - self.duration.max()):
            warnings.warn(
                "half-sine pulse extends beyond the bounds of the time series"
            )

        t = np.arange(tstart, tstop, dt)

        data = np.zeros((len(t), len(self.amplitude)), dtype=float)
        t_data, ampl_data, T_data = np.broadcast_arrays(
            t[..., None], self.amplitude.to_numpy(), self.duration.to_numpy()
        )
        t_mask = np.nonzero((t_data >= tpulse) & (t_data < tpulse + T_data))
        data[t_mask] = ampl_data[t_mask] * np.sin(
            np.pi * t_data[t_mask] / T_data[t_mask]
        )

        return pd.DataFrame(
            data,
            index=pd.Series(t, name="timestamp"),
            columns=self.amplitude.index,
        )

    # def widened_duration(self, new_duration: float):
    #    pass

    # def pseudo_velocity(self):
    #    pass


def enveloping_half_sine(
    pvss: pd.DataFrame,
    damp: float = 0.0,
) -> HalfSineWavePulse:
    """
    Characterize a half-sine pulse whose PVSS envelopes the input.

    :param pvss: the PVSS to envelope
    :param damp: the damping factor used to generate the input PVSS
    :return: a tuple of amplitudes and periods, each pair of which describes a
        half-sine pulse

    .. seealso::

        `Pseudo Velocity Shock Spectrum Rules For Analysis Of Mechanical Shock, Howard A. Gaberson <https://info.endaq.com/hubfs/pvsrs_rules.pdf>`_
    """

    def amp_factor(damp):
        """
        Calculate the PVSS amplitude attenuation on a half-sine pulse from the
        damping coefficient.

        The PVSS of a half-sine pulse differs based on the damping coefficient
        used. While the high-frequency rolloff is relatively consistent, the
        flat low-frequency amplitude is attenuated at higher damping values.
        This function calculates this attenuation for a given damping
        coefficient.
        """
        # This calculates the PVSS value as ω->0. However, since it necessarily
        # computes the maximum of a function *over time*, and ω is only found
        # therein in the multiplicative factor (ωt), it is mathematically
        # equivalent to compute this maximum for any arbitrary ω>0. Thus we
        # choose ω=1 for convenience, w/o loss of generality.
        a = np.exp(1j * np.arccos(-damp))  # = -damp + 1j * np.sqrt(1 - damp**2)
        # From WolframAlpha: https://www.wolframalpha.com/input/?i=D%5BPower%5Be%2C%5C%2840%29-d+*t%5C%2841%29%5D+sin%5C%2840%29Sqrt%5B1-Power%5Bd%2C2%5D%5D*t%5C%2841%29%2Ct%5D+%3D+0&assumption=%22ListOrTimes%22+-%3E+%22Times%22&assumption=%7B%22C%22%2C+%22e%22%7D+-%3E+%7B%22NamedConstant%22%7D&assumption=%7B%22C%22%2C+%22d%22%7D+-%3E+%7B%22Variable%22%7D&assumption=%22UnitClash%22+-%3E+%7B%22d%22%2C+%7B%22Days%22%7D%7D
        t_max = (2 / np.imag(a)) * np.arctan2(np.imag(a), 1 - np.real(a))
        PVSS_max = (1 / np.imag(a)) * np.imag(np.exp(a * t_max))
        return PVSS_max

    max_pvss = pvss.max()
    max_f_pvss = pvss.mul(pvss.index, axis=0).max()

    return HalfSineWavePulse(
        amplitude=2 * np.pi * max_f_pvss,
        duration=max_pvss / (4 * amp_factor(damp) * max_f_pvss),
    )



def single_file_analysis_script(files: list,
                                file_download_url: str,
                                channel: str = '40g DC Acceleration',
                                subchannel: str = 'Z (40g)') -> str:
    """
    Creates a cloud dashboard (for use in cloud.endaq.com custom report generation) for plotting and exploring the
    data in the most recently uploaded file to the enDAQ Cloud.  It creates a PSD plot with octave spaced frequency
    bins, a plot of the channel data around the peak value, a spectrogram of subchannel data that's frequency bins are
    octave spaced, and a rolling envelope of every sensors data.

    :param files: A `list` of JSON blobs of recording file data as it's given in a custom report in cloud.endaq.com
    :param file_download_url: The download URL for the IDE file most recently uploaded to cloud.endaq.com
    :param channel: The name of the channel to focus the analysis on
    :param subchannel: The name of the subchannel to focus some of the analysis on
    :return: The json based string which is to be given as the variable 'output' in enDAQ cloud
     custom dashboards (to produce the dashboard)
    """
    from endaq.ide import get_doc, to_pandas
    from endaq.plot import octave_spectrogram, octave_psd_bar_plot, around_peak
    from endaq.plot.dashboards import rolling_enveloped_dashboard
    from endaq.plot.utilities import set_theme
    from endaq.cloud import create_cloud_dashboard_output

    # Set the aesthetic theme of the figures (defaults to enDAQ dark theme)
    set_theme()

    # Gets the IDE data as an `idelib.dataset.Dataset` object
    doc = get_doc(file_download_url)

    # A dictionary mapping channel names to pandas dataframes of that channels data
    channel_dict = {doc.channels[ch].name: to_pandas(doc.channels[ch]) for ch in doc.channels}

    # Check that the desired channel exists
    if channel not in channel_dict:
        raise ValueError(f"'{channel}' was not in the list of available channels.")

    # Check that the desired subchannel exists
    if subchannel not in channel_dict[channel]:
        raise ValueError(f"'{subchannel}' was not in the list of available subchannels of {channel}.")

    # Get the acceleration dataframe
    acceleration_df = channel_dict[channel]

    name_to_fig = dict()

    # Create a bar plot of PSD data with octave spaced frequency bins
    name_to_fig["Acceleration PSD"] = octave_psd_bar_plot(acceleration_df[[subchannel]], yaxis_title=subchannel)

    # Create a plot of the data surrounding the time of peak magnitude in the channel
    name_to_fig["Acceleration Around Peak"] = around_peak(acceleration_df, 1000)

    # Create a spectrogram with octave spaced frequency bins
    name_to_fig["Z-Axis Acceleration Spectrogram"] = octave_spectrogram(acceleration_df[[subchannel]], window=.15)[-1]

    # Generate Row of Subplots
    name_to_fig["Rolling Envelopes"] = rolling_enveloped_dashboard(
        channel_dict, plot_as_bars=True, num_rows=1, num_cols=None)

    # Create JSON Output
    output = create_cloud_dashboard_output(name_to_fig)

    return output



import codecs
import os.path
import setuptools


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

INSTALL_REQUIRES = [
    "backports.cached-property; python_version<'3.8'",
    "ebmlite>=3.2.0",
    "idelib>=3.2.3",
    "jinja2",
    "numpy>=1.19.5",
    "pandas>=1.3",
    "plotly>=5.3.1",
    "python-dotenv>=0.18.0",
    "requests>=2.25.1",
    "scipy>=1.7.1",
    "pint>=0.18"
    ]

TEST_REQUIRES = [
    "hypothesis==6.41.0",
    "pytest",
    "pytest-cov",
    "pytest-xdist[psutil]",
    "sympy",
    ]

DOCS_REQUIRES = [
    "sphinx",
    "pydata-sphinx-theme",
    "sphinx-plotly-directive",
    "sphinxcontrib-spelling",
    "nbsphinx",
    ]

EXAMPLE_REQUIRES = [
    ]

setuptools.setup(
        name='endaq',
        version=get_version('endaq/__init__.py'),
        author='Mide Technology',
        author_email='help@mide.com',
        description='A comprehensive, user-centric Python API for working with enDAQ data and devices',
        long_description=long_description,
        long_description_content_type='text/markdown',
        url='https://github.com/MideTechnology/endaq-python',
        license='MIT',
        classifiers=['Development Status :: 4 - Beta',
                     'License :: OSI Approved :: MIT License',
                     'Natural Language :: English',
                     'Programming Language :: Python :: 3.7',
                     'Programming Language :: Python :: 3.8',
                     'Programming Language :: Python :: 3.9',
                     'Topic :: Scientific/Engineering',
                     ],
        keywords='ebml binary ide mide endaq',
        packages=setuptools.find_packages(exclude=('tests',)),
        package_dir={'endaq': './endaq'},
        project_urls={
            "Bug Tracker": "https://github.com/MideTechnology/endaq-python/issues",
            "Documentation": "https://docs.endaq.com/en/latest/",
            "Source Code": "https://github.com/MideTechnology/endaq-python/tree/main",
            },
        install_requires=INSTALL_REQUIRES,
        extras_require={
            'test': INSTALL_REQUIRES + TEST_REQUIRES,
            'docs': INSTALL_REQUIRES + DOCS_REQUIRES,
            'example': INSTALL_REQUIRES + EXAMPLE_REQUIRES,
            },
)
